from rdflib import Graph, Literal, URIRef, Namespace, Dataset, BNode
import csv
from rdflib.plugins import sparql
from rdflib.namespace import XSD, RDF, FOAF 
import re
from itertools import takewhile
import os
import sys, getopt
import urllib.parse
from datetime import datetime
import click


def rr(vocab):
	""" return r2rml vocabulary item """
	rr = u'http://www.w3.org/ns/r2rml#'	
	return rr + vocab
	
class CsvOptions:
	""" Class to represent CSV file options """
	def __init__(self, encoding="utf-8", delimiter=",", dateformat='%Y-%m-%d'):
		self.delimiter = delimiter
		self.encoding = encoding
		self.dateformat = dateformat
		
class CsvInfo:
	""" Handles the opening and reading of csv files """ 
	def __init__(self, path, options = None):
		self.path = path
		self.columnNames = None
		self.columnTypes = None
		self.csvfile = None
		self.reader = None
		self.errors = []
		self.tablename = os.path.splitext(os.path.basename(self.path))[0]
		self.options = options
		if not options:
			self.options = CsvOptions()
			
	def __enter__(self):
		self.reset_file()
		return self
		
	def reset_file(self):
		self.csvfile = open(self.path, encoding=self.options.encoding)
		self.reader = csv.DictReader(self.csvfile,delimiter=self.options.delimiter)
		self.set_column_names()
		return self
		
	def __exit__(self, *args):
		if self.csvfile:
			self.csvfile.close()
			
			
	def get_restarted_reader(self):
		if self.csvfile.closed:
			self.reset_file()
		self.csvfile.seek(0)
		return self.reader
		
		
	def set_column_names(self):
		rdr = self.get_restarted_reader()
		try:
			self.columnNames = [name for name in next(rdr)]
		except StopIteration as e:
			self.errors.append('csv file {} may be empty'.format(self.tablename))
			
	def get_contents(self):
		reader = self.get_restarted_reader()
		buf = []
		for line in reader:
			buf.append(line)
			
		return buf
		
class Rml:
	""" Reads config mapping file and applies mappings to the data in one or more csvfiles"""
	def __init__(self):
		# graph to hold triplemap elements
		self.g = Graph()
		
		# graph to hold generated triples
		self.triples = Graph()
		
		# @baseURL of rml file
		self._baseURL = ''
		
		# self.making_quads is true if rml has any triplesmaps that name a non 'defaultGraph'
		self.making_quads = False
		
		self.csv_paths = []
		self.csvInfoList = []
		self.csv_tables = []
		self.errors = []
		
	def loadFile(self, file, csv_arr = [], options = None):
		self.csv_paths = csv_arr
		self.options = options
		self.open_csvs()
		self.g.parse(file, format="ttl")
		self.set_baseURI(file)
		return len(self.g)
		
		
	def open_csvs(self, options = None):
		for csvfile in self.csv_paths:
			with CsvInfo(csvfile, self.options) as csvInfo:
				self.csvInfoList.append(csvInfo)
				self.csv_tables.append(csvInfo.tablename)
				# append errors
				for err in csvInfo.errors:
					self.errors.append(err)
					
					
	# get the csvinfo that matches this tablename
	def get_csvInfo_by_tablename(self, tablename):
		for csvInfo in self.csvInfoList:
			if csvInfo.tablename == tablename:
				return csvInfo
		return None
		
		
	def set_baseURI(self,path):
		with open(path) as file:
			# an entry starts with `#@` line and ends with a blank line
			
			for line in file:
				if line.startswith('@base'):
					buf = [line]
					#       read until blank line
					buf.extend(takewhile(str.strip, file)) #
			#       if just one return it
			if len(buf) == 1:
				self.base= re.findall(r'<([^>]+)>', ''.join(buf))[0]
				
	
	def get_baseURI(self):
		return self.base
		
	def runSparql(self,qstring):
		base = Namespace(URIRef(self.base))
		
		rr = Namespace("http://www.w3.org/ns/r2rml#")
		#self.g.bind("base", base)
		
		self.g.namespace_manager.bind('', base)
		q = sparql.prepareQuery(qstring,initNs= {"" : base,"xsd" : XSD, "rdf" : RDF, "rr" : rr})
		
		return self.g.query(q)
		

	def tmaps(self):
		tmaps = {}
		# for each triplemap add a list of dict for predicateObjectMaps called 'poms'
		for tm, p, o in self.g.triples((None, RDF.type, URIRef(rr('TriplesMap')))):
			tmaps[str(tm)] = {
				'poms' : [],
				rr('logicalTable') : {},
				rr('subjectMap') : {}
				}
			# for each predicateObjectMap find its bnode obj
			for tmap, p, bn in self.g.triples((tm, URIRef(rr('predicateObjectMap')), None)):
				pom = {}
				# for each pom's bnode as subject 
				for bnk, pompred, pomobj in self.g.triples((bn, None, None)):
					if not isinstance(pomobj, BNode):
						# if we already have a predicate append obj to list
						if str(pompred) in pom and len(pom[str(pompred)]) > 0:
							pom[str(pompred)] += ',' + str(pomobj)
						else:
							pom[str(pompred)] = str(pomobj)
					else:
						pom[str(pompred)] = {}
						for pmbkn, pmbknp, pmbkno in self.g.triples((pomobj,None,None)):
							if not isinstance(pmbkno, BNode):
								pom[str(pompred)][str(pmbknp)] = str(pmbkno).strip('\""')
							else:
								pom[str(pompred)][str(pmbknp)] = {}
								for js, jp, jo in self.g.triples((pmbkno,None,None)):
									pom[str(pompred)][str(pmbknp)][str(jp)] = str(jo).strip('\""')
								
				tmaps[str(tm)]['poms'].append(pom)
			# use logicalTable's blank node to find all contents
			logtable_bnode = self.g.value(subject = tm, predicate = URIRef(rr('logicalTable')))
			
			for ltbn, ltpred, ltobj in self.g.triples((logtable_bnode, None, None)):
				tmaps[str(tm)][rr('logicalTable')][str(ltpred)] = str(ltobj).strip('\""')
				
			# use subjectMap's blank node to find all contents
			#smap_bnode = self.g.value(subject = tm, predicate = rr('subjectMap'))
			smap_bnode = None
			for x, y, z in self.g.triples((tm, URIRef(rr('subjectMap')), None)):
				if isinstance(z, BNode):
					smap_bnode = z
			
			# may get many obj with same pred e.g. rr:class
			for smbn, smpred, smobj in self.g.triples((smap_bnode, None, None)):
				if not str(smpred) in tmaps[str(tm)][rr('subjectMap')]:
					tmaps[str(tm)][rr('subjectMap')][str(smpred)] = ""
					
				if str(smpred) == rr('class'):
					if len(tmaps[str(tm)][rr('subjectMap')][rr('class')]) > 0:
						tmaps[str(tm)][rr('subjectMap')][rr('class')] += ',' + str(smobj).strip('\""')
					else:
						tmaps[str(tm)][rr('subjectMap')][rr('class')] = str(smobj).strip('\""')
				elif str(smpred) ==  rr('template') and not isinstance(smobj,BNode):
					tmaps[str(tm)][rr('subjectMap')][rr('template')] = str(smobj).strip('\""')
					#if 'graph' in str(smobj).strip('\""'):
						#print('smoobj=' + str(smobj).strip('\""'))
				elif str(smpred) ==  rr('graph'):
					tmaps[str(tm)][rr('subjectMap')][rr('graph')] = str(smobj).strip('\""')
				elif str(smpred) == rr('graphMap'):
					tmaps[str(tm)][rr('subjectMap')][rr('graphMap')] = {}
					for smbkn, smbknp, smbnkno in self.g.triples((smobj,None,None)):					
						tmaps[str(tm)][rr('subjectMap')][rr('graphMap')][str(smbknp)] = str(smbnkno).strip('\""')
		
		for rom, x, y in self.g.triples((None, RDF.type, URIRef(rr('RefObjectMap')))):
			tmaps[str(rom)] = {}
			for refom, ptm, tmref in self.g.triples((rom, URIRef(rr('parentTriplesMap')), None)):
				tmaps[str(rom)][str(ptm)] = str(tmref)
				
		
		return tmaps
	
	def get_Uri_From_Template(self, template, row):
		cols = []
		s = re.sub('[{"}]','', template)
		for match in re.finditer(r'{"([^>]+?)"}', template):
			cols.append(match.group(0).strip('{}"'))
		for col in cols:
			s = s.replace(col,urllib.parse.quote(row[col]))
		return str(s)
		
	
	# use base and the tmap structure populated from the rml file and its named columns in the csv/table named by tablename to create triples in  the triples graph
	def create_triples(self):
	
		tmaps = self.tmaps()
		
		making_quads = False
		
		# for each triple map in rml
		for tk in tmaps.keys():
			
			# skip the ones without subjectMap; they are RefObjectMaps
			if rr('subjectMap') not in tmaps[tk]:
				continue
			
			# determine if we are making quads or triples and get context if quads
			context = ''
			if rr('graph') in tmaps[tk][rr('subjectMap')]:
				context = tmaps[tk][rr('subjectMap')][rr('graph')]
				if not str(context).endswith('defaultGraph'):
					making_quads = True
					self.making_quads = True
				else:
					making_quads = False
				
				
			elif rr('graphMap') in tmaps[tk][rr('subjectMap')]:
					if rr('column') in tmaps[tk][rr('subjectMap')][rr('graphMap')] and rr('termType') in tmaps[tk][rr('subjectMap')][rr('graphMap')] and tmaps[tk][rr('subjectMap')][rr('graphMap')][rr('termType')] == rr('Literal'):
						# cant use a literal as graph uri
						making_quads = False
						self.errors.append('Graph specified should not be a literal value')
						continue
					elif rr('template') in tmaps[tk][rr('subjectMap')][rr('graphMap')]:
						making_quads = True
						self.making_quads = True
						context = tmaps[tk][rr('subjectMap')][rr('graphMap')][rr('template')]
			else:
				making_quads = False
				
			# reinitialise graph as Dataset as it will be holding quads 	
			if making_quads:
				self.triples = Dataset()
			
			# find col names used in this tmap's template
			cols = []
			subjmap_template = tmaps[tk][rr('subjectMap')][rr('template')]
			for match in re.finditer(r'{"([^>]+?)"}',subjmap_template):
				cols.append(match.group(0).strip('{}"'))
				
			
			# use tablename to look in right csv
			logical_table = ''
			csvinfo = None
			if rr('logicalTable') in tmaps[tk] and rr('tableName') in tmaps[tk][rr('logicalTable')]:
				logical_table = tmaps[tk][rr('logicalTable')][rr('tableName')]
			if logical_table:
				csvinfo = self.get_csvInfo_by_tablename(logical_table)		
				
				#print('logical table: ' + logical_table)
				#if csvinfo is not None:
				#	print('found csvinfo: ' + csvinfo.tablename)
			
			# check we have a csv file with the same name as the logical_table in rml
			if csvinfo is None:
				self.errors.append('A csv for table {} was not found'.format(logical_table))
				continue
				
			
			# only continue if fieldnames in cols named of template all appear in column names of table in csv
			if set(cols).issubset(csvinfo.columnNames):
				# construct URIRef for subject
				rdr = csvinfo.get_restarted_reader()
				
				
				next(rdr) # skip the headers
				for row in rdr:				
					# if subjectmap's graph given by template, col names by row values
					if rr('graphMap') in tmaps[tk][rr('subjectMap')]:
						if rr('template') in tmaps[tk][rr('subjectMap')][rr('graphMap')]:
							context = self.get_Uri_From_Template(context,row)
					
					# replace subject template's col names with values from csv
					s = re.sub('[{"}]','',subjmap_template)
					
					for col in cols:			
						s = s.replace(col,urllib.parse.quote(str(row[col])))
					subject = URIRef(s)
					object = None
					predicate = None
					# create triples / quads for each type in class list from rml file
					csv_classes = ''
					if rr('class') in tmaps[tk][rr('subjectMap')]:
						csv_classes = tmaps[tk][rr('subjectMap')][rr('class')]
						classes = csv_classes.split(',')
						for cls in classes:
							predicate = RDF.type
							object = URIRef(cls)
							
							if making_quads:
								self.triples.add((subject, predicate, object,URIRef(context)))
							else:
								self.triples.add((subject, predicate, object))
					
					# iterate through predobjs 
					for pom in tmaps[tk]['poms']:
						predicates = str(pom[rr('predicate')]).split(',')
						for pred_str in predicates:
							predicate = URIRef(pred_str)
						
							pom_column = ''
							pom_object = ''
							if rr('object') in pom:
								pom_object = pom[rr('object')]
							if rr('objectMap') in pom and rr('column') in pom[rr('objectMap')]:
								pom_column = pom[rr('objectMap')][rr('column')]
								
							# if objectmap in pom and key of this om is also in tmap then it is a refobjmap
							if rr('objectMap') in pom and str(pom[rr('objectMap')]) in tmaps.keys() and rr('parentTriplesMap') in tmaps[str(pom[rr('objectMap')])] and tmaps[str(pom[rr('objectMap')])][rr('parentTriplesMap')] in tmaps.keys():
								rom = pom[rr('objectMap')]
								
								ptm = tmaps[rom][rr('parentTriplesMap')]
								ptmSubjectTemplate = tmaps[ptm][rr('subjectMap')][rr('template')]
								object = URIRef(self.get_Uri_From_Template(ptmSubjectTemplate,row))
								
							elif rr('objectMap') in pom and rr('parentTriplesMap') in pom[rr('objectMap')]:
								rom = pom[rr('objectMap')]
								
								ptm = rom[rr('parentTriplesMap')]
								
								if ptm in tmaps.keys():
									if rr('joinCondition') in rom:
										join = rom[rr('joinCondition')]
										if rr('child') in join and rr('parent') in join:
											child = join[rr('child')]
											childval = row[child]
											# if null or empty no triple to make fir the fk relationship
											if childval is None or str(childval).strip() == '':
												continue
											
											parent = join[rr('parent')]
											
											# get ref tmap's logical table to get csv, get row in that where col parent val is childval
											fklogicaltable = tmaps[ptm][rr('logicalTable')][rr('tableName')]
											fkcsvinfo = self.get_csvInfo_by_tablename(fklogicaltable)
											fkrdr = fkcsvinfo.get_restarted_reader()
											
											next(fkrdr) # skip the headers
											record = None
											
											for fkrow in fkrdr:
												if fkrow[parent] ==  childval:
													record = fkrow
											# get uri from fk csv's template and fk row
											ptmSubjectTemplate = tmaps[ptm][rr('subjectMap')][rr('template')]
											
											object = URIRef(self.get_Uri_From_Template(ptmSubjectTemplate,record))
							
							elif len(pom_column) > 0 and pom_column in row.keys():
								
								object = self.get_literal(row[pom_column],csvinfo.options.dateformat)
							else:
								object = URIRef(pom_object)
	
							# add to Dataset for quads if non-default graph else add just add triple
							if making_quads:
								gc = self.triples.graph(URIRef(context))
								self.triples.add((subject, predicate, object, gc))
								
							else:
								self.triples.add((subject, predicate, object))
			else:
				# cols mentioned in template that are not cols in csv
				self.errors.append('Fields are named in a template that are not columns in a CSV file.')
		# add every namespace from rml file to outputted triple/quad file
		for n in self.g.namespace_manager.namespaces():
			self.triples.namespace_manager.bind(n[0],n[1])
		self.triples.namespace_manager.bind('rdf', RDF)	
	
	def get_literal(self, value, dateformat='%Y-%m-%d'):
		
		try:
			dt = datetime.strptime(value,dateformat).date()
			return Literal(dt, datatype=XSD.date)
		except ValueError:
			pass
		try: 
			i = int(value)
			return Literal(i, datatype=XSD.integer)
		except ValueError:
			pass
		try:
			f = float(value)
			return Literal(f, datatype=XSD.float)
		except ValueError:
			pass
			
		return Literal(value)


	def write_file(self, outfile, format="n3"):
		self.triples.serialize(destination=outfile, format=format)
				
										
	def serialise(self):
		return self.triples.serialize(format="turtle").decode("utf-8")
		
@click.command()
@click.option("--map", "-m", "mapfile", required=True,
				type=click.Path(exists=True),
            	help="Path to the Triples Map configuration file")
@click.option("--csvfile", "-c", "csvfile", required=True,
            	type=click.Path(exists=True),
            	help="Path to each CSV file. This option can be specified for each CSV file.",
            	multiple=True)
@click.option("--outfile", "-o", "outfile", required=True, 
				type=click.Path(),
				help="Path to the output file. It should have '.ttl' file type for triples or a '.nq' for nquads when you have specified a named graph other than 'defaultGraph' for the 'rr:graph' setting in the configuration file.")
@click.option("--separator", "-s",
              help="The CSV separator or delimiter. Defaults to comma.",
              default=",")
@click.option("--encoding", "-e",
              help="The input CSV's file encoding. Defaults to 'utf-8'.",
              default="utf-8")
@click.option("--dateformat", "-d",
              help="If you have dates in your CSV files this describes the format they are in. See 'strptime format codes'. Defaults to '%Y-%m-%d', for example 2022-01-31.",     
              default="%Y-%m-%d")
def process(mapfile, csvfile, outfile, separator=',', encoding='utf-8', dateformat='%Y-%m-%d'):
	""" Generates triples or n-quads in outfile from one or more CSV files according to the configuration in the Triples Map configuration file. """
	csvfiles = list(csvfile)
	if not sys.stdin.isatty():
		csvfiles.extend(list(sys.stdin))
		
	s = Rml()
	
	# take option defaults or those entered
	options = CsvOptions(encoding = encoding, delimiter = separator, dateformat = dateformat)
	
	# load one rml and 1 or more csvs
	s.loadFile(mapfile, csvfiles, options)
	s.create_triples()
	try:
	    s.write_file(outfile)
	except Exception as err:
	    print("Error outputting graph: {0}".format(err))
	    raise
	for e in s.errors:
		print(e)
	
if __name__ == "__main__":
	process()




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['triplify_csv']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0', 'pytest>=6.2.5,<7.0.0', 'rdflib>=6.0.2,<7.0.0']

entry_points = \
{'console_scripts': ['triplify_csv = triplify_csv.triplify_csv:process']}

setup_kwargs = {
    'name': 'triplify-csv',
    'version': '0.1.0',
    'description': 'A tool to generate triples from CSV files according to a configuration file.',
    'long_description': '============\ntriplify_csv\n============\nFeatures\n--------\nGenerates RDF triples or quads in Turtle or NQuad syntax from one or more CSV files and a configuration file.\n\nInstallation\n------------\ntriplify_csv can be installed from PyPI using \'pip\':\n\n.. code-block:: bash\n\n\tpip install triplify_csv\n\n\nUsage\n-----\ntriplify_csv installs as both a package and command line interface tool\n\n**Example of using the package**\n\n.. code-block:: python\n\n\tfrom TriplifyCsv import Rml, CsvOptions\n\t\n\t# config mapping files are .ttl files\n\tconfigfile = \'myconfig.ttl\'\n\t\n\t# csv files should be .csv files\n\tcsvfile1, csvfile2 = \'mycsv1.csv\', \'mycsv2.csv\'\n\t\n\t# output file must have either a .ttl extension for turtle triples \n\t# or a .nq extension for quads\n\toutputfile = \'mytriples.ttl\'\n\t\n\trml = Rml()\n\t\n\t# default date format of dates in your CSV files is \'%Y-%m-%d\'\n\t# default csv delimiter is \',\'\n\t# override the defaults by setting options\n\toptions = CsvOptions(dateformat=\'%d/%m/%Y\', delimiter=\'|\')\n\t\n\t# load one rml and one or more csvs\n\trml.loadFile(configfile, [csvfile1,csvfile2], options)\n\t\n\trml.create_triples()\n\t\n\t# "nquads" for named graphs need a .nq extension\n\t# here we are generating triples so .ttl for turtle syntax\n\trml.write_file(outputfile, format="ttl")\n\t \n\n**Example of CLI use - help text**\n\nTo display full help text on the options enter the following at the command line\n\n.. code-block:: bash\n\t\n\ttriplify_csv --help\n\t \n\n**Example of CLI use - making triples**\nThe same example as the one in code above as a CLI call instead ...\n\n.. code-block:: bash\n\t\n\ttriplify_csv -m \'myconfig.ttl\' -c \'mycsv1.csv\' -c \'mycsv2.csv\' -o \'mytriples.ttl\'\n\t \n**How to make your configuration file**\n\nThe configuration file contains a set of mappings for triplify_csv to follow to set the subjects, predicates and objects or literal values of your triples or nquads from the data in one or more CSV files. These mappings are RDF triples in the turtle syntax.  The terms that can be used are a subset of the terms defined in the R2RML standard.  \n\nR2RML was not designed for this purpose. R2RML is \'.. a language for expressing customized mappings from relational databases to RDF datasets.\' (see https://www.w3.org/TR/r2rml/ ). Triplify_csv uses a subset of R2RML to express customised mappings from CSV files to RDF datasets.  Where R2RML refers to the tables of a database using \'rr:logicalTable\' this should be understood in the triplify_csv use of R2RML as referring to the name (without \'.csv\') of a corresponding csv file.  \'rr:sqlQuery\', the term of the R2RML language that lets you express mappings from database queries to RDF isn\'t supported in the triplify_csv usage.  Also, there is no need to support \'rr:sqlVersion\'.\n\nFor a complete list of what parts of the R2RML language are supported see the examples in the /tests folder and refer to the R2RML test cases document (https://www.w3.org/TR/rdb2rdf-test-cases/). As of version 0.1.0 the test cases supported are\n\n- R2RMLTC0007a - Typing resources by relying on rdf:type predicate\n- R2RMLTC0007b - Assigning triples to Named Graphs\n- R2RMLTC0007c - One column mapping, using rr:class\n- R2RMLTC0007d - One column mapping, specifying an rr:predicateObjectMap with rdf:type\n- R2RMLTC0007e - One column mapping, using rr:graphMap and rr:class\n- R2RMLTC0007f - One column mapping, using rr:graphMap and specifying an rr:predicateObjectMap with rdf:type\n- R2RMLTC0007g - Assigning triples to the default graph\n- R2RMLTC0007h - Assigning triples to a non-IRI named graph\n- R2RMLTC0008a - Generation of triples to a target graph by using rr:graphMap and rr:template\n- R2RMLTC0008b - Generation of triples referencing object map\n- R2RMLTC0008c - Generation of triples by using multiple predicateMaps within a rr:predicateObjectMap\n- R2RMLTC0009a - Generation of triples from foreign key relations\n\n\n**Simple config file example**\nSuppose you have a CSV file containing details of contacts (example format below) and you want to generate RDF data from this using FOAF as the ontology you want to make use of for this, the R2RML config file might look like this ...\n\n.. code-block:: bash\n\n\t@prefix rr: <http://www.w3.org/ns/r2rml#> .\n\t@prefix foaf: <http://xmlns.com/foaf/0.1/> .\n\t@prefix ex: <http://example.com/> .\n\t@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n\t@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n\t@base <http://example.com/base/> .\n\t\n\t<TriplesMap1> a rr:TriplesMap;\n\trr:logicalTable [ rr:tableName "\\"Contacts\\"" ];\n\t\n\trr:subjectMap [ rr:template "http://example.com/Contact/{\\"ID\\"}/{\\"Name\\"}";\n\t rr:class foaf:Person;\n\t];\n\t\n\trr:predicateObjectMap [ rr:predicate ex:id ;\n\t rr:objectMap [ rr:column "\\"ID\\"" ;  ] ;\n\t];\n\t\n\trr:predicateObjectMap [ rr:predicate foaf:name ;\n\t rr:objectMap [ rr:column "\\"Name\\"" ; ] ;\n\t];\n\t\n\trr:predicateObjectMap [ rr:predicate foaf:interest ;\n\t  rr:objectMap [ rr:column "\\"Interest\\"" ; ] ;\n\t];\n\t\n\t.\n\n\n\n\nCreate a CSV file called \'Contacts.csv\' using commas as delimiters between the following values (shown here in a table) ...\n\n.. csv-table:: Contacts.csv\n\t:header: "ID", "Name", "Interest"\n\t:widths: auto\n\n\t10, "John Smith", https://en.m.wikipedia.org/wiki/Tennis\n\t20, "Joe Bloggs", https://en.m.wikipedia.org/wiki/Golf\n\t30, "Mr Bun", https://en.m.wikipedia.org/wiki/Spam_(food)\n\n\nNow, with triplify_csv installed save the R2RML config file as \'contactsmap.ttl\' and the csv file as \'Contacts.csv\' and generate the output containing your triples to a file called \'contactstriples.ttl\' (for example) with the following command ...\n\n.. code-block:: bash\n\t\n\ttriplify_csv -m \'contactsmap.ttl\' -c \'Contacts.csv\' -o \'contactstriples.ttl\'\n\n\t\nThe resulting triples in turtle syntax in the \'contactstriples.ttl\' file\nwould look like this ...\n\n.. code-block:: bash\n\n\t@prefix ex: <http://example.com/> .\n\t@prefix foaf: <http://xmlns.com/foaf/0.1/> .\n\t@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n\t\n\t<http://example.com/Contact/10/John%20Smith> a foaf:Person ;\n\t    ex:id 10 ;\n\t    foaf:interest "https://en.m.wikipedia.org/wiki/Tennis" ;\n\t    foaf:name "John Smith" .\n\t\n\t<http://example.com/Contact/20/Joe%20Bloggs> a foaf:Person ;\n\t    ex:id 20 ;\n\t    foaf:interest "https://en.m.wikipedia.org/wiki/Golf" ;\n\t    foaf:name "Joe Bloggs" .\n\t\n\t<http://example.com/Contact/30/Mr%20Bun> a foaf:Person ;\n\t    ex:id 30 ;\n\t    foaf:interest "https://en.m.wikipedia.org/wiki/Spam_(food)" ;\n\t    foaf:name "Mr Bun" .\n',
    'author': 'Adrian Atley',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
